// run configure with --enable-local
// run make elemlist


#include <click/config.h>
#include "ForwHeader.hh"
#include <click/confparse.hh>
#include <click/error.hh>
#include <list>

#include <iostream>

CLICK_DECLS

ForwHeader::ForwHeader() : PerDestPacketQueue(0), OutPacketQueue(0), _meth(0), _opts(0)
{
}

ForwHeader::~ForwHeader()
{
}

int ForwHeader::configure(Vector<String> &conf, ErrorHandler *errh)
{
  if (cp_va_kparse(conf, this, errh,
		   "HOPS", cpkP, cpIPAddressList, &_hops,
		   "METHOD", cpkP, cpByte, &_meth,
		   "OPTIONS", cpkP, cpByte, &_opts,
		   cpEnd) < 0)
    return -1;

  if(_hops.size() > 256)
      return errh->error("Too many hops %s", _hops.size());
  if(_hops.size() < 2)
      return errh->error("Need at least one hop");

  Vector<IPAddress>::iterator hop=_hops.begin();
  IPAddress dest = *hop; ++hop;
  Vector<IPAddress> route;
  for(;hop != _hops.end();  ++hop)
      route.push_back(*hop);

  _hop_table[dest] = route;

  return 0;
}

void ForwHeader::push_packet(Packet *p, const Vector<IPAddress> &hops)
{
    if(WritablePacket *q = p->push(4 + 4 * hops.size()))
    {
        unsigned char * s = q->data();
        s[0] = (unsigned char)hops.size();
        s[1] = 0;
        s[2] = _meth;
        s[3] = _opts;
        s += 4;
        for(Vector<IPAddress>::const_iterator hop = hops.begin();
            hop != hops.end();
            hop++, s += 4)
            memcpy(s, &(*hop), 4);
              
        output(0).push(q);
    }
}

void ForwHeader::push(int port, Packet *p)
{
  assert(port==0);

  const click_ip *iph = p -> ip_header();
  assert(p->has_network_header());
  const IPAddress addr = iph->ip_dst.s_addr;
  const Vector <IPAddress> &hops = _hop_table.get(addr);
  
  // destination IP not in list

  if(hops.size() == 0)
  {
      struct PacketList *pl = new struct PacketList;
      pl->next = PerDestPacketQueue[addr];
      pl->p = p;
      PerDestPacketQueue[addr] = pl;
  }
  else
      push_packet(p, hops);
}

String ForwHeader::read_handler_queued(Element *e, void *)
{
    String r;
    for(TPerDestPacketQueue::const_iterator it = static_cast<ForwHeader*>(e)->PerDestPacketQueue.begin();
        it;
        ++it)
        if(it.value())
            r = r + it.key().unparse() + " ";
    return r;
}

String ForwHeader::read_handler_dump_routes(Element *e, void *)
{
    String r;
    for(HopTable::const_iterator it = static_cast<ForwHeader*>(e)->_hop_table.begin();
        it;
        ++it)
    {
        r = r + it.key().unparse() + "   ";
        for(Vector<IPAddress>::const_iterator it2 = it.value().begin(); 
            it2 != it.value().end(); 
            ++it2)
            r = r + (*it2).unparse() + " ";
        r = r + "\n";
    }
    return r;
}

int ForwHeader::write_handler_set_route(const String & s, Element *e, void*, ErrorHandler *errh)
{
    IPAddress *ip = 0;
    
    Vector<IPAddress> l;

    const char *begin = 0;

    for(const char *f = s.data(); f <= s.end(); ++f)
        if((f != s.end()) && (*f != ' '))
        {
            if(!begin)
                begin = f;
        }
        else
        {
            if(begin)
            {
                if(!ip)
                    ip = new IPAddress(s.substring(begin-s.data(),f-begin));
                else
                    l.push_back(IPAddress(s.substring(begin-s.data(),f-begin)));
                begin = 0;
            }
        };
                
    if(!ip || !l.size())
        return errh->error("write handler set_route in ForwHeader");

    ForwHeader * const FwdHd = static_cast<ForwHeader*>(e);

    FwdHd->_hop_table[*ip] = l;

    for(struct PacketList* pl = FwdHd->PerDestPacketQueue[*ip]; pl; )
    {
        FwdHd->push_packet(pl -> p, l);
        struct PacketList *next = pl -> next;
        delete pl;
        pl = next;
    }

    delete ip;

    return 0;
}

int ForwHeader::write_handler_clear(const String &, Element *e, void *, ErrorHandler*)
{
    static_cast<ForwHeader*>(e)->_hop_table.clear();
    return 0;
}

void ForwHeader::add_handlers()
{
    add_read_handler("queued", read_handler_queued, (void*)0);
    add_read_handler("dump_routes", read_handler_dump_routes, (void*)0);
    add_write_handler("set_route", write_handler_set_route, (void*)0);
    add_write_handler("clear", write_handler_clear, (void*)0, Handler::BUTTON);

}

CLICK_ENDDECLS
EXPORT_ELEMENT(ForwHeader)
