#ifndef CLICK_FORWHEADER_HH
#define CLICK_FORWHEADER_HH
#include <click/element.hh>
#include <click/ipaddress.hh>
#include <click/hashtable.hh>
CLICK_DECLS
/*
=c 
ForwHeader(HOPS, METHOD, OPTIONS)

=s Forwarding Network

encapsulates IP packets in Forwarding Header with the hop list being looked up 
in a routing table maintained by ForwHeader

=d

encapsulates each packet in the Forwarding Header. The METHOD and OPTIONS 
8-bit fields in the Forwarding Header will be set with the values of METHOD and
OPTIONS. The hop list written into the Forwarding Header will be looked up in a
routing table maintained by ForwHeader. If no entry for the destination IP is
found, the packet will be enqueued until an entry for the destination IP is
added.

HOPS will be used to initialize the routing table with one row. The first IP is
the destination, the remaining IPs the hop list. This is for testing purposes,
the preferred way to set up the routing table is through the handlers.

=e 
ForwHeader(192.168.151.1 192.168.152.1 192.168.175.1,1,10)

will initialize the routing table so that IP packets with destination 
192.168.151.1 are encapsulated in a Forwarding Header with hop list 
192.168.152.1 192.168.175.1.

=h set_route write-only

Sets the hop list for a destination in the routing table. Expects a list of IP
addresses, the first one is the destination, the remaining ones the hop list.

=h clear write-only

Clears the routing table

=h dump_routes read-only

Lists the routing table. Each row first lists the destination IP, then the hop
list.

=h queued read-only

Lists the destination IPs of all IP packets which are queued because their
destination IP was not in the routing table.

=a ForwHeaderPrint, NextHop, RevRoute
*/

class ForwHeader : public Element {
public:
  ForwHeader();
  ~ForwHeader();
  const char *class_name() const { return "ForwHeader"; }
  int configure(Vector<String> &, ErrorHandler *);
  const char *port_count() const { return PORTS_1_1; }
  const char *processing() const { return PUSH; }
  void push(int port, Packet *p);
  void add_handlers();
  static int write_handler_set_route(const String &, Element *e, void *thunk, ErrorHandler *errh);
  static int write_handler_clear(const String &, Element *e, void *thunk, ErrorHandler *errh);
  static String read_handler_queued(Element *e, void *thunk);
  static String read_handler_dump_routes(Element *e, void *thunk);
  
  struct PacketList {
        Packet* p;
        struct PacketList* next;
  };
  
  typedef HashTable<IPAddress,Vector<IPAddress> > HopTable;
  typedef HashTable<IPAddress, struct PacketList*> TPerDestPacketQueue;

private:
  HopTable _hop_table;
  Vector<IPAddress> _hops;
  TPerDestPacketQueue PerDestPacketQueue;
  struct PacketList* OutPacketQueue;
  void push_packet(Packet *p, const Vector<IPAddress> &hops);

  unsigned char _meth;
  unsigned char _opts;
};

CLICK_ENDDECLS
#endif
