// -*- c-basic-offset: 4 -*-
/*
 * vsystun.{cc,hh} -- element accesses network via tun device allocated 
 * through vsys
 * Matthias Goerner
 *
 * based on kerneltun.{cc,hh}
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

#include <click/config.h>
#include "vsystun.hh"
#include <click/error.hh>
#include <click/bitvector.hh>
#include <click/confparse.hh>
#include <click/straccum.hh>
#include <click/glue.hh>
#include <clicknet/ether.h>
#include <click/standard/scheduleinfo.hh>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>

#include <sys/un.h>
#include <errno.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/if_tun.h>

#include "elements/userlevel/fakepcap.hh"

CLICK_DECLS

#define VSYS_TUNTAP "/vsys/fd_tuntap.control"

/* Reads vif FD from "fd", writes interface name to vif_name, and returns vif FD.
 * vif_name should be IFNAMSIZ chars long. */
int receive_vif_fd(int fd, char *vif_name)
{
	struct msghdr msg;
	struct iovec iov;
	int rv;
	size_t ccmsg[CMSG_SPACE(sizeof(int)) / sizeof(size_t)];
	struct cmsghdr *cmsg;

    /* Use IOV to read interface name */
	iov.iov_base = vif_name;
	iov.iov_len = IFNAMSIZ;

	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	/* old BSD implementations should use msg_accrights instead of
	 * msg_control; the interface is different. */
	msg.msg_control = ccmsg;
	msg.msg_controllen = sizeof(ccmsg);

	while(((rv = recvmsg(fd, &msg, 0)) == -1) && errno == EINTR) {};
	if (rv == -1) {
		perror("vsystun.cc, receive_vif_fd: recvmsg");
		return -1;
	}
	if(!rv) {
		/* EOF */
		return -1;
	}

	cmsg = CMSG_FIRSTHDR(&msg);
	if (!cmsg->cmsg_type == SCM_RIGHTS) {
		fprintf(stderr, "vsystun.cc, receive_vif_fd: got control message of unknown type %d\n",
			cmsg->cmsg_type);
		return -1;
	}
	return *(int*)CMSG_DATA(cmsg);
}


int tun_alloc(int iftype, char *if_name)
{
    int control_fd;
    struct sockaddr_un addr;
    int remotefd;

    control_fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (control_fd == -1) {
        perror("vsystun.cc, tun_alloc: Could not create UNIX socket\n");
        return -1;
    }

    memset(&addr, 0, sizeof(struct sockaddr_un));
    /* Clear structure */
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, VSYS_TUNTAP,
            sizeof(addr.sun_path) - 1);

    if (connect(control_fd, (struct sockaddr *) &addr,
                sizeof(struct sockaddr_un)) == -1) {
        perror("vsystun.cc, tun_alloc: Could not connect to Vsys control socket");
        return -1;
    }

    /* passing type param */
    if (send(control_fd, &iftype, sizeof(iftype), 0) != sizeof(iftype)) {
        perror("vsystun.cc, tun_alloc: Could not send paramater to Vsys control socket");
        return -1;
    }

    remotefd = receive_vif_fd(control_fd, if_name);
    return remotefd;
}


VsysTun::VsysTun()
    : _fd(-1),
      _task(this), _ignore_q_errs(false),
      _printed_write_err(false), _printed_read_err(false)
{
}

VsysTun::~VsysTun()
{
}

void *
VsysTun::cast(const char *n)
{
    if (strcmp(n, "VsysTun") == 0)
	return this;
    else
	return Element::cast(n);
}

int
VsysTun::configure(Vector<String> &conf, ErrorHandler *errh)
{
    _headroom = Packet::default_headroom;
    _adjust_headroom = false;
    _headroom += (4 - _headroom % 4) % 4; // default 4/0 alignment
    _mtu_out = DEFAULT_MTU;
    if (cp_va_kparse(conf, this, errh,
		     "ADDR", cpkP+cpkM, cpIPPrefix, &_near, &_mask,
                     "VSYS_FILE", 0, cpString, &_vsys_file,
		     "HEADROOM", cpkC, &_adjust_headroom, cpUnsigned, &_headroom,
		     "IGNORE_QUEUE_OVERFLOWS", 0, cpBool, &_ignore_q_errs,
		     "MTU", 0, cpInteger, &_mtu_out,
                     cpEnd) < 0)
	return -1;

    if (_mtu_out < (int) sizeof(click_ip))
	return errh->error("MTU must be greater than %d", sizeof(click_ip));
    if (_headroom > 8192)
	return errh->error("HEADROOM too big");
    _adjust_headroom = !_adjust_headroom;
    return 0;
}

int
VsysTun::setup_tun(ErrorHandler *errh)
{

    char pref_len[16];
    char read_buffer[1024];
    sprintf(pref_len,"%d",_mask.mask_to_prefix_len());

    String output = _dev_name + "\n" + 
                    _near.unparse() + "\n" + 
                    pref_len + "\n" +
                    "snat=1\n";
    String out_file = _vsys_file + ".out";
    String in_file = _vsys_file + ".in";

    String chatter_message = "Sending to " + out_file + ":\n" + output;

    click_chatter(chatter_message.c_str());
    
    int fd_in = open(in_file.c_str(), O_WRONLY);
    if(fd_in < 0)
    {
        perror(("While opening file "+in_file).c_str());
        return errh->error("setup_tun in vsystun.cc failed");
    }
    
    int fd_out = open(out_file.c_str(), O_RDONLY);
    if(fd_out < 0)
    {
        perror(("While opening file "+out_file).c_str());
        return errh->error("setup_tun in vsystun.cc failed");
    }

    ssize_t bytes_written = write(fd_in, output.c_str(), strlen(output.c_str()));

    if(bytes_written != (ssize_t)strlen(output.c_str()))
    {
        perror(("While writing to file "+in_file).c_str());
        return errh->error("setup_tun in vsystun.cc failed");
    }
    close(fd_in);

    ssize_t bytes_read = read(fd_out, read_buffer, 1024);

    if(bytes_read < 0)
    {
        perror(("While reading from file "+out_file).c_str());
        return errh->error("setup_tun in vsystun.cc failed");
    }

    if(bytes_read > 0)
    {
        read_buffer[bytes_read]=0;
        fprintf(stderr,"Output at %s: %s",out_file.c_str(),read_buffer);
        return errh->error("setup_tun in vsystun.cc failed");
    }

    close(fd_out);

    _mtu_in = _mtu_out + 4;

    return 0;
}

int
VsysTun::initialize(ErrorHandler *errh)
{
    char if_name[IFNAMSIZ];

    _fd = tun_alloc(IFF_TUN,if_name);
    
    if(_fd < 0)
        return errh->error("Could not open tun through vsys");

    char buffer[1024];
    sprintf(buffer,
            "vsystun, tun_alloc returned file descriptor: %d, device name: %s",
            _fd, if_name);
    click_chatter(buffer);
    
    _dev_name = if_name;

    if (setup_tun(errh) < 0)
         return -1;

    if (input_is_pull(0)) {
	ScheduleInfo::join_scheduler(this, &_task, errh);
	_signal = Notifier::upstream_empty_signal(this, 0, &_task);
    }
    if (_adjust_headroom) {
        _headroom += (4 - _headroom % 4) % 4; // default 4/0 alignment
    }
    add_select(_fd, SELECT_READ);
    return 0;
}

void
VsysTun::cleanup(CleanupStage)
{
    if (_fd >= 0) {
	close(_fd);
	remove_select(_fd, SELECT_READ);
    }
}

void
VsysTun::selected(int fd)
{
    if (fd != _fd)
	return;
    WritablePacket *p = Packet::make(_headroom, 0, _mtu_in, 0);
    if (!p) {
	click_chatter("out of memory!");
	return;
    }

    int cc = read(_fd, p->data(), _mtu_in);
    if (cc > 0) {
	p->take(_mtu_in - cc);
	bool ok = false;

        // Like KernelTun with LINUX_UNIVERSAL

        // 2-byte padding followed by an Ethernet type
        uint16_t etype = *(uint16_t *)(p->data() + 2);
        p->pull(4);
        if (etype != htons(ETHERTYPE_IP) && etype != htons(ETHERTYPE_IP6))
            checked_output_push(1, p->clone());
        else
            ok = fake_pcap_force_ip(p, FAKE_DLT_RAW);

	if (ok) {
	    p->timestamp_anno().assign_now();
	    output(0).push(p);
	} else
	    checked_output_push(1, p);

    } else {
	if (!_ignore_q_errs || !_printed_read_err || (errno != ENOBUFS)) {
	    _printed_read_err = true;
	    perror("VsysTun read");
	}
    }
}

bool
VsysTun::run_task(Task *)
{
    Packet *p = input(0).pull();
    if (p)
	push(0, p);
    else if (!_signal)
	return false;
    _task.fast_reschedule();
    return p != 0;
}

void
VsysTun::push(int, Packet *p)
{
    const click_ip *iph = 0;
    int check_length;

    {
	iph = p->ip_header();
	// check IP header
	if (!p->has_network_header()
	    || p->network_length() < (int) sizeof(click_ip)) {
	    click_chatter("%s(%s): no network header", class_name(), _dev_name.c_str());
	kill:
	    p->kill();
	    return;
	} else if (iph->ip_v != 4 && iph->ip_v != 6) {
	    click_chatter("%s(%s): unknown IP version %d", class_name(), _dev_name.c_str(), iph->ip_v);
	    goto kill;
	}
	// strip link headers
	p->change_headroom_and_length(p->headroom() + p->network_header_offset(), p->network_length());
	check_length = p->length();
    }

    // check MTU
    if (check_length > _mtu_out) {
	click_chatter("%s(%s): packet larger than MTU (%d)", class_name(), _dev_name.c_str(), _mtu_out);
	goto kill;
    }

    WritablePacket *q;
    // Like KernelTun for LINUX_UNIVERSAL
    // 2-byte padding followed by an Ethernet type
    uint32_t ethertype = (iph->ip_v == 4 ? htonl(ETHERTYPE_IP) : htonl(ETHERTYPE_IP6));
    if ((q = p->push(4)))
        *(uint32_t *)(q->data()) = ethertype;
    p = q;

    if (p) {
	int w = write(_fd, p->data(), p->length());
	if (w != (int) p->length() && (errno != ENOBUFS || !_ignore_q_errs || !_printed_write_err)) {
	    _printed_write_err = true;
	    click_chatter("%s(%s): write failed: %s", class_name(), _dev_name.c_str(), strerror(errno));
	}
	p->kill();
    } else
	click_chatter("%s(%s): out of memory", class_name(), _dev_name.c_str());
}

String
VsysTun::print_dev_name(Element *e, void *)
{
    VsysTun *kt = (VsysTun *) e;
    return kt->_dev_name;
}

void
VsysTun::add_handlers()
{
    if (input_is_pull(0))
	add_task_handlers(&_task);
    add_read_handler("dev_name", print_dev_name, 0);
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(userlevel FakePcap)
EXPORT_ELEMENT(VsysTun)
