// -*- c-basic-offset: 4 -*-
#ifndef CLICK_VSYSTUN_HH
#define CLICK_VSYSTUN_HH
#include <click/element.hh>
#include <click/etheraddress.hh>
#include <click/task.hh>
#include <click/notifier.hh>
CLICK_DECLS

/*
=c

VsysTun(ADDR/MASK, VSYS_FILE [, I<keywords> HEADROOM, MTU, IGNORE_QUEUE_OVERFLOWS])

=s comm

Uses planetlab's /vsys/fd_tuntap.control to allocate a tun device. Uses the
specified file to bring up the device on a planetlab node. 

=d

Reads IP packets from and writes IP packets to a tun device reservered through
vsys.

Opens a UNIX socket to /vsys/fd_tuntap.control and calls recvmsg.
vsys will create a tun device and open it and return to click the network
device name and a file descriptor.
click will send the device name, the IP address and prefix length and "snat=1" to
VSYS_FILE.in and read VSYS_FILE.out which is expected to return nothing.
This will bring up the network device and set its IP and also enable source NAT.

To use it on planetlab, you need to have reserved your own private IP and 
vsys access to vif_up and fd_tuntap. After initialization, VsysTun behaves like
KernelTun.

=e 

VsysTun(10.6.0.1/24, VSYS_FILE "/vsys/vif_up")

will set up a tun device with IP 10.6.0.1. An IP packet send to the device
should have source IP 10.6.0.2.

Keyword arguments are

=over 8

=item HEADROOM

Integer. The number of bytes left empty before the packet data to leave
room for additional encapsulation headers. Default is 28.

=item MTU

Integer. The interface's MTU, not including any link headers. KernelTun will
refuse to send packets larger than the MTU. Default is 1500; not all operating
systems allow MTU to be set.

=item IGNORE_QUEUE_OVERFLOWS

Boolean.  If true, don't print more than one error message when
there are queue overflows error when sending/receiving packets
to/from the tun device (e.g. there was an ENOBUFS error).  Default
is false.

=back

=n

This is specfic for planetlab.

=a

KernelTun */

class VsysTun : public Element { public:

    VsysTun();
    ~VsysTun();

    const char *class_name() const	{ return "VsysTun"; }
    const char *port_count() const	{ return "0-1/1-2"; }
    const char *processing() const	{ return "a/h"; }
    const char *flow_code() const	{ return "x/y"; }
    const char *flags() const		{ return "S3"; }

    void *cast(const char *);
    int configure_phase() const		{ return CONFIGURE_PHASE_PRIVILEGED - 1; }
    int configure(Vector<String> &, ErrorHandler *);
    int initialize(ErrorHandler *);
    void cleanup(CleanupStage);
    void add_handlers();

    void selected(int fd);

    void push(int port, Packet *);
    bool run_task(Task *);

  private:

    enum { DEFAULT_MTU = 1500 };

    int _fd;
    int _mtu_in;
    int _mtu_out;
    String _dev_name;
    IPAddress _near;
    IPAddress _mask;
    String _vsys_file;
    unsigned _headroom;
    Task _task;
    NotifierSignal _signal;

    bool _ignore_q_errs;
    bool _printed_write_err;
    bool _printed_read_err;
    bool _adjust_headroom;

    static String print_dev_name(Element *e, void *);

    int setup_tun(ErrorHandler *);
    int updown(IPAddress, IPAddress, ErrorHandler *);

};

CLICK_ENDDECLS
#endif
